"""Example to show dashboard configuration specified as a dictionary."""
import os

import pandas as pd

import vizro.plotly.express as px
from vizro import Vizro
from vizro.actions import export_data, filter_interaction
from vizro.managers import data_manager
from vizro.models import Dashboard


def retrieve_gapminder():
    """This is a function that returns gapminder data."""
    return px.data.gapminder()


def retrieve_gapminder_year(year: int):
    """This is a function that returns gapminder data for a year."""
    return px.data.gapminder().query(f"year == {year}")


def retrieve_gapminder_continent_comparison():
    """This is a function adds aggregated continent information to gapminder data."""
    df_gapminder = px.data.gapminder()
    df_gapminder_agg = px.data.gapminder()

    df_gapminder_agg["lifeExp"] = df_gapminder_agg.groupby(by=["continent", "year"])["lifeExp"].transform("mean")
    df_gapminder_agg["gdpPercap"] = df_gapminder_agg.groupby(by=["continent", "year"])["gdpPercap"].transform("mean")
    df_gapminder_agg["pop"] = df_gapminder_agg.groupby(by=["continent", "year"])["pop"].transform("sum")

    df_gapminder["data"] = "Country"
    df_gapminder_agg["data"] = "Continent"

    df_gapminder_comp = pd.concat([df_gapminder_agg, df_gapminder], ignore_index=True)

    return df_gapminder_comp


def retrieve_avg_gapminder():
    """This is a function that returns aggregated gapminder data."""
    df = px.data.gapminder()
    mean = (
        df.groupby(by=["continent", "year"]).agg({"lifeExp": "mean", "pop": "mean", "gdpPercap": "mean"}).reset_index()
    )
    return mean


def retrieve_avg_gapminder_year(year: int):
    """This is a function that returns aggregated gapminder data for a specific year."""
    return retrieve_avg_gapminder().query(f"year == {year}")


# If you're not interested in lazy loading then you could just do data_manager["gapminder"] = px.data.gapminder()
data_manager["gapminder"] = retrieve_gapminder
data_manager["gapminder_2007"] = lambda: retrieve_gapminder_year(2007)
data_manager["gapminder_avg"] = retrieve_avg_gapminder
data_manager["gapminder_avg_2007"] = lambda: retrieve_avg_gapminder_year(2007)
data_manager["gapminder_country_analysis"] = retrieve_gapminder_continent_comparison

page_variable = {
    "title": "Variable Analysis",
    "layout": {
        "grid": [
            # fmt: off
            [0, 1, 1, 1],
            [2, 3, 3, 3],
            [4, 5, 5, 5],
            [6, 7, 7, 7],
            # fmt: on,
        ],
        "row_min_height": "400px",
        "row_gap": "25px",
    },
    "components": [
        {
            "type": "card",
            "text": """
                ### Overview
                The world map provides initial insights into the variations of metrics across countries and
                continents. Click on Play to see the animation and explore the development over time.

                #### Observation
                A global trend of increasing life expectancy emerges, with some exceptions in specific African
                countries. Additionally, despite similar population growth rates across continents, the overall
                global population continues to expand, with India and China leading the way.  Meanwhile, GDP per
                capita experiences growth in most regions.
            """,
        },
        {
            "type": "graph",
            "id": "variable_map",
            "figure": px.choropleth(
                "gapminder",
                locations="iso_alpha",
                color="lifeExp",
                hover_name="country",
                animation_frame="year",
                labels={
                    "year": "year",
                    "lifeExp": "Life expectancy",
                    "pop": "Population",
                    "gdpPercap": "GDP per capita",
                },
                title="Global development over time",
            ),
        },
        {
            "type": "card",
            "text": """
                    ### Distribution
                    The boxplot illustrates the distribution of each metric across continents, facilitating comparisons
                    of life expectancy, GDP per capita, and population statistics.

                    #### Observation
                    Observations reveal that Europe and Oceania have the highest life expectancy and GDP per capita,
                    likely influenced by their smaller population growth. Additionally, Asia and America exhibit
                    notable GDP per capita outliers, indicating variations among countries within these continents or
                    large growth over the observed years.
            """,
        },
        {
            "type": "graph",
            "id": "variable_boxplot",
            "figure": px.box(
                "gapminder",
                x="continent",
                y="lifeExp",
                color="continent",
                labels={
                    "year": "year",
                    "lifeExp": "Life expectancy",
                    "pop": "Population",
                    "gdpPercap": "GDP per capita",
                    "continent": "Continent",
                },
                title="Distribution per continent",
                color_discrete_map={
                    "Africa": "#00b4ff",
                    "Americas": "#ff9222",
                    "Asia": "#3949ab",
                    "Europe": "#ff5267",
                    "Oceania": "#08bdba",
                },
            ),
        },
        {
            "type": "card",
            "text": """
                    ### Development
                    The line chart tracks the variable's progress from 1952 to 2007, facilitating a deeper comprehension
                    of each metric.

                    #### Observation
                    Oceania and Europe are found to have the highest total GDP per capita and exhibit significant
                    growth. In contrast, Asia, Africa, and America demonstrate a more pronounced upward trend in
                    population increase compared to Europe and Oceania, suggesting that GDP per capita growth might be
                    influenced by relatively smaller population growth in the latter two continents.
            """,
        },
        {
            "type": "graph",
            "id": "variable_line",
            "figure": px.line(
                "gapminder_avg",
                y="lifeExp",
                x="year",
                color="continent",
                title="Development between 1952 and 2007",
                labels={
                    "year": "Year",
                    "lifeExp": "Life expectancy",
                    "pop": "Population",
                    "gdpPercap": "GDP per capita",
                    "continent": "Continent",
                },
                color_discrete_map={
                    "Africa": "#00b4ff",
                    "Americas": "#ff9222",
                    "Asia": "#3949ab",
                    "Europe": "#ff5267",
                    "Oceania": "#08bdba",
                },
            ),
        },
        {
            "type": "card",
            "text": """
                    ### Recent status
                    Examining the data for 2007 provides insight into the current status of each continent and metrics.

                    #### Observation
                    Asia held the largest population, followed by America, Europe, Africa, and Oceania. Life expectancy
                    surpassed 70 years for all continents, except Africa with 55 years. GDP per capita aligns with
                    earlier findings, with Oceania and Europe reporting the highest values and Africa recording the
                    lowest.
                """,
        },
        {
            "type": "graph",
            "id": "variable_bar",
            "figure": px.bar(
                "gapminder_avg_2007",
                x="lifeExp",
                y="continent",
                orientation="h",
                title="Comparison of average metric for 2007",
                labels={
                    "year": "year",
                    "continent": "Continent",
                    "lifeExp": "Life expectancy",
                    "pop": "Population",
                    "gdpPercap": "GDP per capita",
                },
                color="continent",
                color_discrete_map={
                    "Africa": "#00b4ff",
                    "Americas": "#ff9222",
                    "Asia": "#3949ab",
                    "Europe": "#ff5267",
                    "Oceania": "#08bdba",
                },
            ),
        },
    ],
    "controls": [
        {
            "type": "parameter",
            "targets": ["variable_map.color", "variable_boxplot.y", "variable_line.y", "variable_bar.x"],
            "selector": {"type": "radio_items", "title": "Select variable", "options": ["lifeExp", "pop", "gdpPercap"]},
        },
    ],
}

page_relation = {
    "title": "Relationship Analysis",
    "layout": {
        "grid": [[0, 0, 0, 0, 1]] + [[2, 2, 3, 3, 3]] * 4 + [[4, 4, 4, 4, 4]] * 5,
        "row_min_height": "100px",
        "row_gap": "24px",
    },
    "components": [
        {
            "type": "card",
            "text": """
                    Population, GDP per capita, and life expectancy are interconnected metrics that provide insights
                    into the socio-economic well-being of a country.
                    Rapid population growth can strain resources and infrastructure, impacting GDP per capita. Higher
                    GDP per capita often enables better healthcare and improved life expectancy, but other factors such
                    as healthcare quality and social policies also play significant roles.
            """,
        },
        {
            "type": "card",
            "text": """
                    #### Last updated
                    July, 2023
             """,
        },
        {
            "type": "graph",
            "id": "bar_relation_2007",
            "figure": px.box(
                "gapminder_2007",
                x="continent",
                y="lifeExp",
                color="continent",
                hover_name="continent",
                title="Relationship in 2007",
                labels={
                    "gdpPercap": "GDP per capita",
                    "pop": "Population",
                    "lifeExp": "Life expectancy",
                    "continent": "Continent",
                },
                color_discrete_map={
                    "Africa": "#00b4ff",
                    "Americas": "#ff9222",
                    "Asia": "#3949ab",
                    "Europe": "#ff5267",
                    "Oceania": "#08bdba",
                },
                custom_data=["continent"],
            ),
            "actions": [
                {"function": filter_interaction(targets=["scatter_relation_2007"])},
            ],
        },
        {
            "type": "graph",
            "id": "scatter_relation_2007",
            "figure": px.scatter(
                "gapminder_2007",
                x="gdpPercap",
                y="lifeExp",
                size="pop",
                color="continent",
                hover_name="country",
                size_max=60,
                labels={
                    "gdpPercap": "GDP per capita",
                    "pop": "Population",
                    "lifeExp": "Life expectancy",
                    "continent": "Continent",
                },
                color_discrete_map={
                    "Africa": "#00b4ff",
                    "Americas": "#ff9222",
                    "Asia": "#3949ab",
                    "Europe": "#ff5267",
                    "Oceania": "#08bdba",
                },
            ),
        },
        {
            "type": "graph",
            "id": "scatter_relation",
            "figure": px.scatter(
                "gapminder",
                x="gdpPercap",
                y="lifeExp",
                animation_frame="year",
                animation_group="country",
                size="pop",
                color="continent",
                hover_name="country",
                facet_col="continent",
                labels={
                    "gdpPercap": "GDP per capita",
                    "pop": "Population",
                    "lifeExp": "Life expectancy",
                    "continent": "Continent",
                },
                range_y=[25, 90],
                color_discrete_map={
                    "Africa": "#00b4ff",
                    "Americas": "#ff9222",
                    "Asia": "#3949ab",
                    "Europe": "#ff5267",
                    "Oceania": "#08bdba",
                },
            ),
        },
    ],
    "controls": [
        {
            "type": "parameter",
            "targets": ["scatter_relation_2007.x", "scatter_relation.x"],
            "selector": {
                "type": "dropdown",
                "title": "Choose x-axis",
                "options": ["lifeExp", "gdpPercap", "pop"],
                "multi": False,
                "value": "gdpPercap",
            },
        },
        {
            "type": "parameter",
            "targets": ["scatter_relation_2007.y", "scatter_relation.y", "bar_relation_2007.y"],
            "selector": {
                "type": "dropdown",
                "title": "Choose y-axis",
                "options": ["lifeExp", "gdpPercap", "pop"],
                "multi": False,
                "value": "lifeExp",
            },
        },
        {
            "type": "parameter",
            "targets": ["scatter_relation_2007.size", "scatter_relation.size"],
            "selector": {
                "type": "dropdown",
                "title": "Choose bubble size",
                "options": ["lifeExp", "gdpPercap", "pop"],
                "multi": False,
                "value": "pop",
            },
        },
    ],
}

page_continent = {
    "title": "Continent Summary",
    "layout": {"grid": [[i] for i in range(5)], "row_min_height": "190px", "row_gap": "26px"},
    "components": [
        {
            "type": "card",
            "text": """
                    ### Africa
                    ![](assets/images/continents/africa.svg#my-image)

                    Africa, a diverse and expansive continent, faces both challenges and progress in its socio-economic
                    landscape. In 2007, Africa's GDP per capita was approximately $3,000, reflecting relatively slower
                    growth compared to other continents like Oceania and Europe.

                    However, Africa has shown notable improvements in life expectancy over time, reaching 55 years in
                    2007. Despite these economic disparities, Africa's population has been steadily increasing,
                    reflecting its significant potential for development.
            """,
        },
        {
            "type": "card",
            "text": """
                    ### America
                    ![](assets/images/continents/america.svg#my-image)

                    Comprising North and South America, America represents a region of vast geographical and cultural
                    diversity. In 2007, the continent experienced substantial population growth, with a diverse mix of
                    countries contributing to this expansion.

                    Although its GDP per capita of $11,000 in 2007 exhibited variations across countries, America
                    maintained similar levels to Asia, reflecting its economic significance. With North America
                    generally reporting higher life expectancy compared to South America, America remains a region of
                    opportunities and challenges.
            """,
        },
        {
            "type": "card",
            "text": """
                    ### Asia
                    ![](assets/images/continents/asia.svg#my-image)

                    Asia holds a central role in the global economy. It's growth in GDP per capita to $12,000 in 2007
                    and population has been significant, outpacing many other continents. In 2007, it boasted the
                    highest population among all continents, with countries like China and India leading the way.

                    Despite facing various socio-economic challenges, Asia's increasing life expectancy from 46 years
                    to 70 over the years reflects advancements in healthcare and overall well-being, making it a vital
                    region driving global progress and development.
            """,
        },
        {
            "type": "card",
            "text": """
                    ### Europe
                    ![](assets/images/continents/europe.svg#my-image)

                    Europe boasts a strong and thriving economy. In 2007, it exhibited the second-highest GDP per
                    capita of $25,000 among continents, indicating sustained economic growth and development.

                    Europe's life expectancy surpassed 75 years, showcasing a high standard of living and
                    well-established healthcare systems. With its robust infrastructure, advanced industries, and
                    quality of life, Europe continues to be a leading force in the global economy. Between 1952 and
                    2007, Europe's population experienced moderate growth, with a factor of approximately 1.5,
                    notably lower compared to other continents like Asia and America.
            """,
        },
        {
            "type": "card",
            "text": """
                    ### Oceania
                    ![](assets/images/continents/oceania.svg#my-image)

                    Oceania, comprising countries like Australia and New Zealand, stands out with notable economic
                    prosperity and longer life expectancy. In 2007, it boasted the highest GDP per capita of $27,000
                    among continents and exhibited one of the highest life expectancy levels, surpassing 80 years.

                    Despite a relatively smaller population size, Oceania's strong economic growth has contributed
                    to improved living standards and overall well-being of its population.
            """,
        },
    ],
}

page_country = {
    "title": "Country Analysis",
    "layout": {"grid": [[0, 0, 0, 1, 1, 1]] * 7 + [[2, 2, 2, 2, 2, 2]]},
    "components": [
        {
            "type": "graph",
            "id": "bar_country",
            "figure": px.bar(
                "gapminder_country_analysis",
                x="year",
                y="pop",
                color="data",
                barmode="group",
                labels={"year": "Year", "data": "Data", "pop": "Population"},
                color_discrete_map={"Country": "#afe7f9", "Continent": "#003875"},
            ),
        },
        {
            "type": "graph",
            "id": "line_country",
            "figure": px.line(
                "gapminder_country_analysis",
                x="year",
                y="gdpPercap",
                color="data",
                labels={"year": "Year", "data": "Data", "gdpPercap": "GDP per capita"},
                color_discrete_map={"Country": "#afe7f9", "Continent": "#003875"},
                markers=True,
            ),
        },
        {
            "type": "button",
            "id": "export_data_button",
            "text": "Export data",
            "actions": [
                {"function": export_data(targets=["scatter_relation_2007"])},
            ],
        },
    ],
    "controls": [
        {
            "type": "filter",
            "column": "country",
            "selector": {"type": "dropdown", "title": "Select country", "multi": False, "value": "India"},
        },
        {
            "type": "filter",
            "column": "year",
            "selector": {"type": "range_slider", "title": "Select timeframe"},
        },
    ],
}

page_home = {
    "title": "Homepage",
    "layout": {
        "grid": [[0, 1], [2, 3]],
        "col_gap": "24px",
        "row_gap": "16px",
    },
    "components": [
        {
            "type": "card",
            "text": """
                    ![](assets/images/icons/content/hypotheses.svg#icon-top)

                    ### Variable Analysis

                    Analyzing population, GDP per capita and life expectancy on country and continent level.
            """,
            "href": "/variable-analysis",
        },
        {
            "type": "card",
            "text": """
                    ![](assets/images/icons/content/hypotheses.svg#icon-top)

                    ### Relationship Analysis

                    Investigating the interconnection between population, GDP per capita and life expectancy.
            """,
            "href": "/relationship-analysis",
        },
        {
            "type": "card",
            "text": """
                    ![](assets/images/icons/content/collections.svg#icon-top)

                    ### Continent Summary

                    Summarizing the main findings for each continent.
            """,
            "href": "/continent-summary",
        },
        {
            "type": "card",
            "text": """
                    ![](assets/images/icons/content/features.svg#icon-top)

                    ### Country Analysis

                    Discovering how the metrics differ for each country and export data for further investigation.
            """,
            "href": "/country-analysis",
        },
    ],
}

dashboard = {
    "pages": [page_home, page_variable, page_relation, page_continent, page_country],
    "navigation": {
        "pages": {
            "Analysis": ["Homepage", "Variable Analysis", "Relationship Analysis", "Country Analysis"],
            "Summary": ["Continent Summary"],
        }
    },
}
dashboard = Dashboard(**dashboard)

if __name__ == "__main__":
    Vizro._user_assets_folder = os.path.abspath("../assets")
    Vizro().build(dashboard).run()



import pytest
from dash._callback_context import context_value
from dash._utils import AttributeDict

import vizro.models as vm
import vizro.plotly.express as px
from vizro._constants import ON_PAGE_LOAD_ACTION_PREFIX
from vizro._themes import dark, light
from vizro.actions._actions_utils import (
    CallbackTriggerDict,
)
from vizro.managers import model_manager


@pytest.fixture
def target_scatter_filtered_continent_and_pop_parameter_y_and_x(request, gapminder_2007, scatter_params):
    continent, pop, y, x, template = request.param
    continent = continent if isinstance(continent, list) else [continent]
    data = gapminder_2007[
        gapminder_2007["continent"].isin(continent) & gapminder_2007["pop"].between(pop[0], pop[1], inclusive="both")
    ]
    scatter_params["y"] = y
    scatter_params["x"] = x
    return px.scatter(data, template=dark if template == "vizro_dark" else light, **scatter_params).update_layout(
        margin_t=24
    )


@pytest.fixture
def target_box_filtered_continent_and_pop_parameter_y_and_x(request, gapminder_2007, box_params):
    continent, pop, y, x, template = request.param
    continent = continent if isinstance(continent, list) else [continent]
    data = gapminder_2007[
        gapminder_2007["continent"].isin(continent) & gapminder_2007["pop"].between(pop[0], pop[1], inclusive="both")
    ]
    box_params["y"] = y
    box_params["x"] = x
    return px.box(data, template=dark if template == "vizro_dark" else light, **box_params).update_layout(margin_t=24)


@pytest.fixture
def callback_context_on_page_load(request):
    """Mock dash.callback_context that represents on page load."""
    continent_filter, pop, y, x, template = request.param
    mock_callback_context = {
        "args_grouping": {
            "filter_interaction": [],
            "filters": [
                CallbackTriggerDict(
                    id="continent_filter",
                    property="value",
                    value=continent_filter,
                    str_id="continent_filter",
                    triggered=False,
                ),
                CallbackTriggerDict(
                    id="pop_filter",
                    property="value",
                    value=pop,
                    str_id="pop_filter",
                    triggered=False,
                ),
            ],
            "parameters": [
                CallbackTriggerDict(
                    id="y_parameter",
                    property="value",
                    value=y,
                    str_id="y_parameter",
                    triggered=False,
                ),
                CallbackTriggerDict(
                    id="x_parameter",
                    property="value",
                    value=x,
                    str_id="x_parameter",
                    triggered=False,
                ),
            ],
            "theme_selector": CallbackTriggerDict(
                id="theme_selector",
                property="on",
                value=True if template == "vizro_dark" else False,
                str_id="theme_selector",
                triggered=False,
            ),
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


class TestOnPageLoad:
    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_on_page_load, target_scatter_filtered_continent_and_pop_parameter_y_and_x",
        [
            (
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_dark"],
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_dark"],
            ),
            (
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_light"],
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_light"],
            ),
        ],
        indirect=True,
    )
    def test_multiple_controls_one_target(
        self,
        callback_context_on_page_load,
        target_scatter_filtered_continent_and_pop_parameter_y_and_x,
        box_chart,
    ):
        # Creating and adding a Filter objects to the existing Page
        continent_filter = vm.Filter(
            id="test_filter_continent",
            column="continent",
            targets=["scatter_chart"],
            selector=vm.Dropdown(id="continent_filter"),
        )
        pop_filter = vm.Filter(
            id="test_filter_pop", column="pop", targets=["scatter_chart"], selector=vm.RangeSlider(id="pop_filter")
        )
        model_manager["test_page"].controls = [continent_filter, pop_filter]

        # Adds a default _filter Action to the filter selector objects
        continent_filter.pre_build()
        pop_filter.pre_build()

        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            id="test_parameter_y",
            targets=["scatter_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        x_parameter = vm.Parameter(
            id="test_parameter_x",
            targets=["scatter_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [y_parameter, x_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()
        x_parameter.pre_build()

        # Run action by picking 'on_page_load' default Page action function and executing it with ()
        result = model_manager[f"{ON_PAGE_LOAD_ACTION_PREFIX}_action_test_page"].function()

        assert result["scatter_chart"] == target_scatter_filtered_continent_and_pop_parameter_y_and_x

    @pytest.mark.usefixtures("managers_one_page_two_graphs_one_button")
    @pytest.mark.parametrize(
        "callback_context_on_page_load, "
        "target_scatter_filtered_continent_and_pop_parameter_y_and_x, "
        "target_box_filtered_continent_and_pop_parameter_y_and_x",
        [
            (
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_dark"],
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_dark"],
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_dark"],
            ),
            (
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_light"],
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_light"],
                ["Africa", [10**6, 10**7], "pop", "continent", "vizro_light"],
            ),
        ],
        indirect=True,
    )
    def test_multiple_controls_multiple_targets(
        self,
        callback_context_on_page_load,
        target_scatter_filtered_continent_and_pop_parameter_y_and_x,
        target_box_filtered_continent_and_pop_parameter_y_and_x,
    ):
        # Creating and adding a Filter objects to the existing Page
        continent_filter = vm.Filter(
            column="continent", targets=["scatter_chart", "box_chart"], selector=vm.Dropdown(id="continent_filter")
        )
        pop_filter = vm.Filter(
            column="pop", targets=["scatter_chart", "box_chart"], selector=vm.RangeSlider(id="pop_filter")
        )
        model_manager["test_page"].controls = [continent_filter, pop_filter]

        # Adds a default _filter Action to the filter selector objects
        continent_filter.pre_build()
        pop_filter.pre_build()

        # Creating and adding a Parameter object to the existing Page
        y_parameter = vm.Parameter(
            targets=["scatter_chart.y", "box_chart.y"],
            selector=vm.RadioItems(id="y_parameter", options=["lifeExp", "pop", "gdpPercap"], value="lifeExp"),
        )
        x_parameter = vm.Parameter(
            targets=["scatter_chart.x", "box_chart.x"],
            selector=vm.RadioItems(id="x_parameter", options=["continent", "country"], value="continent"),
        )
        model_manager["test_page"].controls = [y_parameter, x_parameter]

        # Adds a default _parameter Action to the "y_parameter" selector object
        y_parameter.pre_build()
        x_parameter.pre_build()

        # Run action by picking 'on_page_load' default Page action function and executing it with ()
        result = model_manager[f"{ON_PAGE_LOAD_ACTION_PREFIX}_action_test_page"].function()

        assert result["scatter_chart"] == target_scatter_filtered_continent_and_pop_parameter_y_and_x
        assert result["box_chart"] == target_box_filtered_continent_and_pop_parameter_y_and_x



"""Unit tests for vizro.models.Card."""
import json

import dash_bootstrap_components as dbc
import plotly
import pytest
from dash import dcc, html
from pydantic import ValidationError

import vizro.models as vm


@pytest.fixture
def expected_card():
    text = dcc.Markdown("Hello", className="card_text", dangerously_allow_html=False, id="card_id")
    button = html.Div(
        dbc.Button(
            href="https://www.google.com",
            className="card_button",
        ),
        className="button_container",
    )

    return html.Div([text, button], className="nav_card_container", id="card_id_outer")


class TestCardInstantiation:
    """Tests model instantiation."""

    def test_create_card_mandatory_only(self):
        card = vm.Card(text="Text to test card")

        assert hasattr(card, "id")
        assert card.type == "card"
        assert card.text == "Text to test card"
        assert card.href is None

    @pytest.mark.parametrize("id, href", [("id_1", "/page_1_reference"), ("id_2", "https://www.google.de/")])
    def test_create_card_mandatory_and_optional(self, id, href):
        card = vm.Card(text="Text to test card", id=id, href=href)

        assert card.type == "card"
        assert card.text == "Text to test card"
        assert card.id == id
        assert card.href == href

    def test_mandatory_text_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Card()

    def test_none_as_text(self):
        with pytest.raises(ValidationError, match="none is not an allowed value"):
            vm.Card(text=None)


class TestBuildMethod:
    """Tests build method."""

    def test_card_build(self, expected_card):
        card = vm.Card(text="Hello", id="card_id", href="https://www.google.com")
        card = card.build()
        result = json.loads(json.dumps(card, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_card, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected

    @pytest.mark.parametrize(
        "test_text, expected",
        [
            ("""# Header""", "# Header"),
            ("""_This **tests** emphasis_""", "_This **tests** emphasis_"),
            ("""> This tests blockquotes""", "> This tests blockquotes"),
            ("""* This tests list items""", "* This tests list items"),
            ("Text to test card", "Text to test card"),
            ("", ""),
            (
                """![](assets/images/icons/content/hypotheses.svg#icon-top)""",
                "![](assets/images/icons/content/hypotheses.svg#icon-top)",
            ),
            ("""Code block: ```python print(1)```""", "Code block: ```python print(1)```"),
            ("""[Example page](/test_page)""", "[Example page](/test_page)"),
        ],
    )
    def test_markdown_setting(self, test_text, expected):
        card = vm.Card(text=test_text, id="id_valid")
        card = card.build()
        card_markdown = card["id_valid"]

        assert isinstance(card_markdown, dcc.Markdown)
        assert card_markdown.dangerously_allow_html is False
        assert card_markdown.children == expected

    @pytest.mark.parametrize(
        "test_text, expected",
        [
            ("""<p>Hello </p>""", "<p>Hello </p>"),  # html will not be evaluated but converted to string
            (12345, "12345"),
            ("""$$ \\frac{1}{(\\sqrt{\\phi \\sqrt{5}}-\\phi)}$$""", "$$ \\frac{1}{(\\sqrt{\\phi \\sqrt{5}}-\\phi)}$$"),
        ],
    )
    def test_markdown_build_invalid(self, test_text, expected):
        card = vm.Card(text=test_text, id="test_id")
        card = card.build()
        card_markdown = card["test_id"]

        assert isinstance(card_markdown, dcc.Markdown)
        assert card_markdown.dangerously_allow_html is False
        assert card_markdown.children == expected



from __future__ import annotations

from typing import List, Optional

import dash_bootstrap_components as dbc
import dash_daq as daq
from dash import Input, Output, Patch, callback, dcc, html
from pydantic import Field, root_validator, validator

import vizro._themes as themes
from vizro._constants import ON_PAGE_LOAD_ACTION_PREFIX
from vizro.actions import _on_page_load
from vizro.managers import model_manager
from vizro.managers._model_manager import DuplicateIDError
from vizro.models import Action, Dashboard, Graph, Layout, VizroBaseModel
from vizro.models._action._actions_chain import ActionsChain, Trigger
from vizro.models._models_utils import _log_call, get_unique_grid_component_ids

from .types import ComponentType, ControlType


class Page(VizroBaseModel):
    """A page in [`Dashboard`][vizro.models.Dashboard] with its own URL path and place in the `Navigation`.

    Args:
        components (List[ComponentType]): See [ComponentType][vizro.models.types.ComponentType]. At least one component
            has to be provided.
        title (str): Title to be displayed.
        layout (Optional[Layout]): Layout to place components in. Defaults to `None`.
        controls (List[ControlType]): See [ControlType][vizro.models.types.ControlType]. Defaults to `[]`.
        path (Optional[str]): Path to navigate to page. Defaults to `None`.

    Raises:
        ValueError: If number of page and grid components is not the same
    """

    components: List[ComponentType]
    title: str = Field(..., description="Title to be displayed.")
    layout: Optional[Layout]
    controls: List[ControlType] = []
    path: Optional[str] = Field(None, description="Path to navigate to page.")

    # TODO: Remove default on page load action if possible
    actions: List[ActionsChain] = []

    @root_validator(pre=True)
    def set_id(cls, values):
        if "title" not in values:
            return values

        values.setdefault("id", values["title"])
        return values

    @validator("components", always=True)
    def set_components(cls, components):
        if not components:
            raise ValueError("Ensure this value has at least 1 item.")
        return components

    @validator("layout", always=True)
    def set_layout(cls, layout, values) -> Layout:
        if "components" not in values:
            return layout

        if layout is None:
            grid = [[i] for i in range(len(values["components"]))]
            return Layout(grid=grid)

        unique_grid_idx = get_unique_grid_component_ids(layout.grid)
        if len(unique_grid_idx) != len(values["components"]):
            raise ValueError("Number of page and grid components need to be the same.")
        return layout

    @validator("path", always=True)
    def set_path(cls, path, values) -> str:
        # Based on how Github generates anchor links - see:
        # https://stackoverflow.com/questions/72536973/how-are-github-markdown-anchor-links-constructed.
        def clean_path(path, allowed_characters):
            path = path.strip().lower().replace(" ", "-")
            path = "".join(character for character in path if character.isalnum() or character in allowed_characters)
            return "/" + path

        # Allow "/" in path if provided by user, otherwise turn page id into suitable URL path (not allowing "/")
        if path:
            return clean_path(path, "-_/")
        return clean_path(values["id"], "-_")

    def __init__(self, **data):
        """Adds the model instance to the model manager."""
        try:
            super().__init__(**data)
        except DuplicateIDError as exc:
            raise ValueError(
                f"Page with id={self.id} already exists. Page id is automatically set to the same "
                f"as the page title. If you have multiple pages with the same title then you must assign a unique id."
            ) from exc

    @_log_call
    def pre_build(self):
        # TODO: Remove default on page load action if possible
        if any(isinstance(component, Graph) for component in self.components):
            self.actions = [
                ActionsChain(
                    id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_{self.id}",
                    trigger=Trigger(
                        component_id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_trigger_{self.id}",
                        component_property="data",
                    ),
                    actions=[
                        Action(
                            id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_action_{self.id}", function=_on_page_load(page_id=self.id)
                        )
                    ],
                )
            ]

    @_log_call
    def build(self):
        self._update_graph_theme()
        controls_content = [control.build() for control in self.controls]
        components_content = [
            html.Div(
                component.build(),
                style={
                    "gridColumn": f"{grid_coord.col_start}/{grid_coord.col_end}",
                    "gridRow": f"{grid_coord.row_start}/{grid_coord.row_end}",
                },
            )
            for component, grid_coord in zip(
                self.components, self.layout.component_grid_lines  # type: ignore[union-attr]
            )
        ]
        return self._make_page_layout(controls_content, components_content)

    def _update_graph_theme(self):
        outputs = [
            Output(component.id, "figure", allow_duplicate=True)
            for component in self.components
            if isinstance(component, Graph)
        ]
        if outputs:

            @callback(
                outputs,
                Input("theme_selector", "on"),
                prevent_initial_call="initial_duplicate",
            )
            def update_graph_theme(theme_selector_on: bool):
                patched_figure = Patch()
                patched_figure["layout"]["template"] = themes.dark if theme_selector_on else themes.light
                return [patched_figure] * len(outputs)

    @staticmethod
    def _create_theme_switch():
        _, dashboard = next(model_manager._items_with_type(Dashboard))
        theme_switch = daq.BooleanSwitch(
            id="theme_selector", on=True if dashboard.theme == "vizro_dark" else False, persistence=True
        )
        return theme_switch

    @staticmethod
    def _create_control_panel(controls_content):
        keyline = html.Div(className="keyline")
        control_panel = html.Div(
            children=[*controls_content, keyline],
            className="control_panel",
        )
        return control_panel if controls_content else None

    def _create_nav_panel(self):
        from vizro.models._navigation._accordion import Accordion

        _, dashboard = next(model_manager._items_with_type(Dashboard))
        if dashboard.navigation:
            return dashboard.navigation.build()

        return Accordion().build()

    def _create_component_container(self, components_content):
        component_container = html.Div(
            children=html.Div(
                components_content,
                style={
                    "gridRowGap": self.layout.row_gap,  # type: ignore[union-attr]
                    "gridColumnGap": self.layout.col_gap,  # type: ignore[union-attr]
                    "gridTemplateColumns": f"repeat({len(self.layout.grid[0])},"  # type: ignore[union-attr]
                    f"minmax({self.layout.col_min_width}, 1fr))",
                    "gridTemplateRows": f"repeat({len(self.layout.grid)},"  # type: ignore[union-attr]
                    f"minmax({self.layout.row_min_height}, 1fr))",
                },
                className="component_container_grid",
            ),
            className="component_container",
        )
        return component_container

    @staticmethod
    def _arrange_containers(page_title, theme_switch, nav_panel, control_panel, component_container):
        """Defines div container arrangement on page.

        To change arrangement, one has to change the order in the header, left_side and/or right_side_elements.
        """
        _, dashboard = next(model_manager._items_with_type(Dashboard))
        dashboard_title = (
            html.Div(
                children=[html.H2(dashboard.title), html.Div(className="keyline")], className="dashboard_title_outer"
            )
            if dashboard.title
            else None
        )

        header_elements = [page_title, theme_switch]
        left_side_elements = [dashboard_title, nav_panel, control_panel]
        header = html.Div(
            children=header_elements,
            className="header",
        )
        left_side = html.Div(children=left_side_elements, className="left_side") if any(left_side_elements) else None
        right_side_elements = [header, component_container]
        right_side = html.Div(
            children=right_side_elements,
            className="right_side",
        )
        return left_side, right_side

    def _make_page_layout(self, controls_content, components_content):
        # Create dashboard containers/elements
        page_title = html.H2(children=self.title)
        theme_switch = self._create_theme_switch()
        nav_panel = self._create_nav_panel()
        control_panel = self._create_control_panel(controls_content)
        component_container = self._create_component_container(components_content)

        # Arrange dashboard containers
        left_side, right_side = self._arrange_containers(
            page_title=page_title,
            theme_switch=theme_switch,
            nav_panel=nav_panel,
            control_panel=control_panel,
            component_container=component_container,
        )

        return dbc.Container(
            id=self.id,
            children=[dcc.Store(id=f"{ON_PAGE_LOAD_ACTION_PREFIX}_trigger_{self.id}"), left_side, right_side],
            className="page_container",
        )



"""Contains utilities for the implementation of action functions."""

from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, TypedDict

import pandas as pd

from vizro._constants import ALL_OPTION
from vizro.managers import data_manager, model_manager
from vizro.managers._model_manager import ModelID
from vizro.models.types import SelectorType

if TYPE_CHECKING:
    from vizro.models import Action


class CallbackTriggerDict(TypedDict):  # shortened as 'ctd'
    id: ModelID  # the component ID. If it`s a pattern matching ID, it will be a dict.
    property: Literal["clickData", "value", "n_clicks"]  #  the component property used in the callback.
    value: Optional[List[Any]]  # the value of the component property at the time the callback was fired.
    str_id: str  # for pattern matching IDs, it`s the stringified dict ID with no white spaces.
    triggered: bool  #  a boolean indicating whether this input triggered the callback.


# Utility functions for helper functions used in pre-defined actions ----
def _get_component_actions(component) -> List[Action]:
    return (
        [action for actions_chain in component.actions for action in actions_chain.actions]
        if hasattr(component, "actions")
        else []
    )


def _apply_filters(
    data_frame: pd.DataFrame,
    ctds_filters: List[CallbackTriggerDict],
    target: str,
) -> pd.DataFrame:
    for ctd in ctds_filters:
        selector_value = ctd["value"]
        selector_value = selector_value if isinstance(selector_value, list) else [selector_value]
        selector_actions = _get_component_actions(model_manager[ctd["id"]])

        for action in selector_actions:
            if target not in action.function["targets"] or ALL_OPTION in selector_value:
                continue

            _filter_function = action.function["filter_function"]
            _filter_column = action.function["filter_column"]
            _filter_value = selector_value
            data_frame = data_frame[_filter_function(data_frame[_filter_column], _filter_value)]

    return data_frame


def _apply_filter_interaction(
    data_frame: pd.DataFrame,
    ctds_filter_interaction: List[CallbackTriggerDict],
    target: str,
) -> pd.DataFrame:
    for ctd in ctds_filter_interaction:
        if not ctd["value"]:
            continue

        source_chart_id = ctd["id"]
        source_chart_actions = _get_component_actions(model_manager[source_chart_id])

        try:
            custom_data_columns = model_manager[source_chart_id]["custom_data"]  # type: ignore[index]
        except KeyError as exc:
            raise KeyError(f"No `custom_data` argument found for source chart with id {source_chart_id}.") from exc

        customdata = ctd["value"]["points"][0]["customdata"]  # type: ignore[call-overload]

        for action in source_chart_actions:
            if target not in action.function["targets"]:
                continue

            for custom_data_idx, column in enumerate(custom_data_columns):
                data_frame = data_frame[data_frame[column].isin([customdata[custom_data_idx]])]
    return data_frame


def _create_target_arg_mapping(dot_separated_strings: List[str]) -> Dict[str, List[str]]:
    results = defaultdict(list)
    for string in dot_separated_strings:
        if "." not in string:
            raise ValueError(f"Provided string {string} must contain a '.'")
        component, arg = string.split(".", 1)
        results[component].append(arg)
    return results


def _update_nested_graph_properties(graph_config: Dict[str, Any], dot_separated_string: str, value: Any):
    keys = dot_separated_string.split(".")
    current_property = graph_config
    for key in keys[:-1]:
        current_property = current_property[key]
    current_property[keys[-1]] = value
    return graph_config


def _get_parametrized_config(targets: List[str], parameters: List[CallbackTriggerDict]) -> Dict[str, Dict[str, Any]]:
    parameterized_config = {}
    for target in targets:
        # TODO - avoid calling _captured_callable. Once we have done this we can remove _arguments from
        #  CapturedCallable entirely.
        graph_config = deepcopy(model_manager[target].figure._arguments)  # type: ignore[index, attr-defined]
        if "data_frame" in graph_config:
            graph_config.pop("data_frame")

        for ctx_trigger_dict in parameters:
            selector_value = ctx_trigger_dict["value"]
            selector_actions = _get_component_actions(model_manager[ctx_trigger_dict["id"]])

            for action in selector_actions:
                action_targets = _create_target_arg_mapping(action.function["targets"])

                if target not in action_targets:
                    continue

                if hasattr(selector_value, "__iter__") and ALL_OPTION in selector_value:  # type: ignore[operator]
                    selector: SelectorType = model_manager[ctx_trigger_dict["id"]]
                    selector_value = selector.options

                for action_targets_arg in action_targets[target]:
                    graph_config = _update_nested_graph_properties(
                        graph_config=graph_config, dot_separated_string=action_targets_arg, value=selector_value
                    )

        parameterized_config[target] = graph_config

    return parameterized_config


# Helper functions used in pre-defined actions ----
def _get_filtered_data(
    targets: List[str],
    ctds_filters: List[CallbackTriggerDict],
    ctds_filter_interaction: List[CallbackTriggerDict],
) -> Dict[str, pd.DataFrame]:
    filtered_data = {}
    for target in targets:
        data_frame = data_manager._get_component_data(target)

        data_frame = _apply_filters(
            data_frame=data_frame,
            ctds_filters=ctds_filters,
            target=target,
        )
        data_frame = _apply_filter_interaction(
            data_frame=data_frame,
            ctds_filter_interaction=ctds_filter_interaction,
            target=target,
        )

        filtered_data[target] = data_frame

    return filtered_data


def _get_modified_page_charts(
    ctds_filter: List[CallbackTriggerDict],
    ctds_filter_interaction: List[CallbackTriggerDict],
    ctds_parameters: List[CallbackTriggerDict],
    ctd_theme: CallbackTriggerDict,
    targets: Optional[List[str]] = None,
) -> Dict[str, Any]:
    if not targets:
        targets = []
    filtered_data = _get_filtered_data(
        targets=targets,
        ctds_filters=ctds_filter,
        ctds_filter_interaction=ctds_filter_interaction,
    )

    parameterized_config = _get_parametrized_config(
        targets=targets,
        parameters=ctds_parameters,
    )

    outputs = {
        target: model_manager[target](  # type: ignore[index, operator]
            data_frame=filtered_data[target],
            **parameterized_config[target],
        ).update_layout(template="vizro_dark" if ctd_theme["value"] else "vizro_light")
        for target in targets
    }

    return outputs



import plotly.express as plotly_express
import plotly.graph_objects as go
import pytest
from pydantic import Field, ValidationError

from vizro.charts._charts_utils import _DashboardReadyFigure
from vizro.models import VizroBaseModel
from vizro.models.types import CapturedCallable, capture


@pytest.fixture
def varargs_function():
    def function(*args, b=2):
        return args[0] + b

    return CapturedCallable(function, 1)


@pytest.mark.xfail
# Known bug: *args doesn't work properly. Fix while keeping the more important test_varkwargs
# passing due to https://bugs.python.org/issue41745.
# Error raised is IndexError: tuple index out of range
def test_varargs(varargs_function):
    assert varargs_function(b=2) == 1 + 2


@pytest.fixture
def positional_only_function():
    def function(a, /, b):
        return a + b

    return CapturedCallable(function, 1)


@pytest.mark.xfail
# Known bug: position-only argument doesn't work properly. Fix while keeping the more important
# test_varkwargs passing due to https://bugs.python.org/issue41745.
# Error raised is TypeError: function got some positional-only arguments passed as keyword arguments: 'a'
def test_positional_only(positional_only_function):
    assert positional_only_function(b=2) == 1 + 2


@pytest.fixture
def keyword_only_function():
    def function(a, *, b):
        return a + b

    return CapturedCallable(function, 1)


def test_keyword_only(keyword_only_function):
    assert keyword_only_function(b=2) == 1 + 2


@pytest.fixture
def varkwargs_function():
    def function(a, b=2, **kwargs):
        return a + b + kwargs["c"]

    return CapturedCallable(function, 1)


def test_varkwargs(varkwargs_function):
    varkwargs_function(c=3, d=4) == 1 + 2 + 3


@pytest.fixture
def simple_function():
    def function(a, b, c, d=4):
        return a + b + c + d

    return CapturedCallable(function, 1, b=2)


class TestCall:
    def test_call_missing_argument(self, simple_function):
        with pytest.raises(TypeError, match="missing 1 required positional argument"):
            simple_function()

    def test_call_needs_keyword_arguments(self, simple_function):
        with pytest.raises(TypeError, match="takes 1 positional argument but 2 were given"):
            simple_function(2)

    def test_call_provide_required_argument(self, simple_function):
        assert simple_function(c=3) == 1 + 2 + 3 + 4

    def test_call_override_existing_arguments(self, simple_function):
        assert simple_function(a=5, b=2, c=6) == 5 + 2 + 6 + 4

    def test_call_is_memoryless(self, simple_function):
        simple_function(c=3)

        with pytest.raises(TypeError, match="missing 1 required positional argument"):
            simple_function()

    def test_call_unknown_argument(self, simple_function):
        with pytest.raises(TypeError, match="got an unexpected keyword argument"):
            simple_function(e=1)


class TestDunderMethods:
    def test_getitem_known_args(self, simple_function):
        assert simple_function["a"] == 1
        assert simple_function["b"] == 2

    def test_getitem_unknown_args(self, simple_function):
        with pytest.raises(KeyError):
            simple_function["c"]

        with pytest.raises(KeyError):
            simple_function["d"]

    def test_delitem(self, simple_function):
        del simple_function["a"]

        with pytest.raises(KeyError):
            simple_function["a"]


def undecorated_function(a, b, c, d=4):
    return a + b + c + d


@capture("action")
def decorated_action_function(a, b, c, d=4):
    return a + b + c + d


@capture("graph")
def decorated_graph_function(data_frame):
    return go.Figure()


@capture("graph")
def decorated_graph_function_px(data_frame):
    return plotly_express.scatter(
        data_frame=data_frame,
        x="gdpPercap",
        y="lifeExp",
    )


@capture("graph")
def invalid_decorated_graph_function():
    return go.Figure()


class Model(VizroBaseModel):
    # The import_path here makes it possible to import the above function using getattr(import_path, _target_).
    function: CapturedCallable = Field(..., import_path=__import__(__name__))


class TestModelFieldPython:
    def test_decorated_action_function(self):
        model = Model(function=decorated_action_function(a=1, b=2))
        assert model.function(c=3, d=4) == 1 + 2 + 3 + 4

    def test_decorated_graph_function(self):
        model = Model(function=decorated_graph_function(data_frame=None))
        assert model.function() == go.Figure()

    def test_undecorated_function(self):
        with pytest.raises(ValidationError, match="provide a valid CapturedCallable object"):
            Model(function=undecorated_function(1, 2, 3))


class TestCapture:
    def test_decorated_df_standard_case(self, gapminder):
        fig = decorated_graph_function_px(gapminder)
        assert len(fig.data) > 0
        assert fig.__class__ == _DashboardReadyFigure

    def test_decorated_df_str(self):
        fig = decorated_graph_function_px("gapminder")
        assert fig == _DashboardReadyFigure()
        assert len(fig.data) == 0

    def test_decorated_graph_function_missing_data_frame(self):
        with pytest.raises(
            ValueError,
            match="decorated_graph_function must supply a value to data_frame argument",
        ):
            decorated_graph_function()

    def test_invalid_decorated_graph_function(self):
        with pytest.raises(
            ValueError,
            match="invalid_decorated_graph_function must have data_frame argument",
        ):
            invalid_decorated_graph_function()


class TestModelFieldJSONConfig:
    def test_no_arguments(self):
        config = {"_target_": "decorated_action_function"}
        model = Model(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function(a=1, b=2, c=3, d=4) == 1 + 2 + 3 + 4

    def test_some_arguments(self):
        config = {"_target_": "decorated_action_function", "a": 1, "b": 2}
        model = Model(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function(c=3) == 1 + 2 + 3 + 4

    def test_all_arguments(self):
        config = {"_target_": "decorated_action_function", "a": 1, "b": 2, "c": 3}
        model = Model(function=config)
        assert isinstance(model.function, CapturedCallable)
        assert model.function() == 1 + 2 + 3 + 4

    def test_decorated_graph_function(self):
        config = {"_target_": "decorated_graph_function", "data_frame": None}
        model = Model(function=config)
        assert model.function() == go.Figure()

    def test_no_target(self):
        config = {"a": 1, "b": 2}
        with pytest.raises(ValidationError, match="must contain the key '_target_'"):
            Model(function=config)

    def test_invalid_import(self):
        config = {"_target_": "invalid_function"}
        with pytest.raises(ValidationError, match="_target_=invalid_function cannot be imported"):
            Model(function=config)

    def test_invalid_arguments(self):
        config = {"_target_": "decorated_action_function", "e": 5}
        with pytest.raises(ValidationError, match="got an unexpected keyword argument"):
            Model(function=config)

    def test_undecorated_function(self):
        config = {"_target_": "undecorated_function", "a": 1, "b": 2, "c": 3}
        with pytest.raises(ValidationError, match="_target_=undecorated_function must be wrapped"):
            Model(function=config)



"""Unit tests for vizro.models.RadioItems."""
import json

import plotly
import pytest
from dash import dcc, html
from pydantic import ValidationError

from vizro.models._action._action import Action
from vizro.models._components.form import RadioItems


@pytest.fixture()
def expected_radio_items():
    return html.Div(
        [
            html.P("Title"),
            dcc.RadioItems(
                id="radio_items_id",
                options=["A", "B", "C"],
                value="A",
                className="selector_body_radio_items",
                persistence=True,
            ),
        ],
        className="selector_container",
        id="radio_items_id_outer",
    )


class TestRadioItemsInstantiation:
    """Tests model instantiation."""

    def test_create_radio_items_mandatory_only(self):
        radio_items = RadioItems()

        assert hasattr(radio_items, "id")
        assert radio_items.type == "radio_items"
        assert radio_items.options == []
        assert radio_items.value is None
        assert radio_items.title is None
        assert radio_items.actions == []

    def test_create_radio_items_mandatory_and_optional(self):
        radio_items = RadioItems(options=["A", "B", "C"], value="A", title="Title", id="radio_items_id")

        assert radio_items.id == "radio_items_id"
        assert radio_items.type == "radio_items"
        assert radio_items.options == ["A", "B", "C"]
        assert radio_items.value == "A"
        assert radio_items.title == "Title"
        assert radio_items.actions == []

    @pytest.mark.parametrize(
        "test_options, expected",
        [
            (["A", "B", "C"], ["A", "B", "C"]),
            ([1, 2, 3], [1.0, 2.0, 3.0]),
            ([1.0, 2.0, 3.0], [1.0, 2.0, 3.0]),
            ([True, False, True], [True, False, True]),
            (
                [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}],
                [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}],
            ),
            (
                [{"label": "New York", "value": "NYC"}, {"label": "Berlin", "value": "BER"}],
                [{"label": "New York", "value": "NYC"}, {"label": "Berlin", "value": "BER"}],
            ),
            (
                [{"label": "True", "value": True}, {"label": "False", "value": False}],
                [{"label": "True", "value": True}, {"label": "False", "value": False}],
            ),
            ([True, 2.0, 1.0, "A", "B"], ["True", "2.0", "1.0", "A", "B"]),
        ],
    )
    def test_create_radio_items_valid_options(self, test_options, expected):
        radio_items = RadioItems(options=test_options)

        assert hasattr(radio_items, "id")
        assert radio_items.type == "radio_items"
        assert radio_items.options == expected
        assert radio_items.value is None
        assert radio_items.title is None
        assert radio_items.actions == []

    @pytest.mark.parametrize("test_options", [1, "A", True, 1.0])
    def test_create_radio_items_invalid_options_type(self, test_options):
        with pytest.raises(ValidationError, match="value is not a valid list"):
            RadioItems(options=test_options)

    def test_create_radio_items_invalid_options_dict(self):
        with pytest.raises(
            ValidationError,
            match="Invalid argument `options` passed. Expected a dict with keys `label` and `value`.",
        ):
            RadioItems(options=[{"hello": "A", "world": "A"}, {"hello": "B", "world": "B"}])

    @pytest.mark.parametrize(
        "test_value, options",
        [
            ("A", ["A", "B", "C"]),
            (1, [1, 2, 3]),
            (2.0, [1.0, 2.0, 3.0]),
            (True, [True, False]),
            ("B", [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
            ("True", [True, 2.0, 1.0, "A", "B"]),
        ],
    )
    def test_create_radio_items_valid_value(self, test_value, options):
        radio_items = RadioItems(options=options, value=test_value)

        assert hasattr(radio_items, "id")
        assert radio_items.type == "radio_items"
        assert radio_items.value == test_value
        assert radio_items.title is None
        assert radio_items.actions == []

    @pytest.mark.parametrize(
        "test_value, options",
        [
            (5, ["A", "B", "C"]),
            (5, [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
            ("D", ["A", "B", "C"]),
            ("D", [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
            (True, ["A", "B", "C"]),
            (True, [{"label": "A", "value": "A"}, {"label": "B", "value": "B"}]),
        ],
    )
    def test_create_radio_items_invalid_value_non_existing(self, test_value, options):
        with pytest.raises(ValidationError, match="Please provide a valid value from `options`."):
            RadioItems(value=test_value, options=options)

    def test_create_radio_items_invalid_value_format(self):
        with pytest.raises(ValidationError, match="3 validation errors for RadioItems"):
            RadioItems(value=[1], options=[1, 2, 3, 4, 5])

    def test_set_action_via_validator(self, test_action_function):
        radio_items = RadioItems(actions=[Action(function=test_action_function)])
        actions_chain = radio_items.actions[0]
        assert actions_chain.trigger.component_property == "value"


class TestRadioItemsBuild:
    """Tests model build method."""

    def test_radio_items_build(self, expected_radio_items):
        radio_items = RadioItems(options=["A", "B", "C"], id="radio_items_id", title="Title").build()

        result = json.loads(json.dumps(radio_items, cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_radio_items, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected



"""Creates action_callback_mapping to map callback arguments to action functions."""

from typing import Any, Dict, Union

from dash import Input, Output, State

from vizro.actions import export_data, filter_interaction
from vizro.actions._callback_mapping._callback_mapping_utils import (
    _get_action_callback_inputs,
    _get_action_callback_outputs,
    _get_export_data_callback_components,
    _get_export_data_callback_outputs,
)
from vizro.actions._filter_action import _filter
from vizro.actions._on_page_load_action import _on_page_load
from vizro.actions._parameter_action import _parameter
from vizro.managers import model_manager
from vizro.managers._model_manager import ModelID


def _get_action_callback_mapping(action_id: ModelID, argument: str) -> Union[Dict[str, Union[Input, State, Output]]]:
    """Creates mapping of action name and required callback input/output."""
    action_function = model_manager[action_id].function._function  # type: ignore[attr-defined]

    action_callback_mapping: Dict[str, Any] = {
        export_data.__wrapped__: {
            "inputs": _get_action_callback_inputs,
            "components": _get_export_data_callback_components,
            "outputs": _get_export_data_callback_outputs,
        },
        _filter.__wrapped__: {
            "inputs": _get_action_callback_inputs,
            "outputs": _get_action_callback_outputs,
        },
        filter_interaction.__wrapped__: {
            "inputs": _get_action_callback_inputs,
            "outputs": _get_action_callback_outputs,
        },
        _parameter.__wrapped__: {
            "inputs": _get_action_callback_inputs,
            "outputs": _get_action_callback_outputs,
        },
        _on_page_load.__wrapped__: {"inputs": _get_action_callback_inputs, "outputs": _get_action_callback_outputs},
    }
    action_call = action_callback_mapping.get(action_function, {}).get(argument, {})
    return action_call if isinstance(action_call, dict) else action_call(action_id=action_id)



"""Pre-defined action function "_on_page_load" to be re-used in `action` parameter of VizroBaseModels."""

from typing import Any, Dict

from dash import ctx

from vizro.actions._actions_utils import (
    _get_modified_page_charts,
)
from vizro.managers import data_manager, model_manager
from vizro.managers._model_manager import ModelID
from vizro.models.types import capture


@capture("action")
def _on_page_load(page_id: ModelID, **inputs: Dict[str, Any]) -> Dict[str, Any]:
    """Applies controls to charts on page once the page is opened (or refreshed).

    Args:
        page_id: Page ID of relevant page
        inputs: Dict mapping action function names with their inputs e.g.
            inputs = {'filters': [], 'parameters': ['gdpPercap'], 'filter_interaction': [], 'theme_selector': True}

    Returns:
        Dict mapping targeted chart ids to modified figures e.g. {'my_scatter': Figure({})}
    """
    targets = [component.id for component in model_manager[page_id].components if data_manager._has_registered_data(component.id)]  # type: ignore[attr-defined]  # noqa: E501

    return _get_modified_page_charts(
        targets=targets,
        ctds_filter=ctx.args_grouping["filters"],
        ctds_filter_interaction=ctx.args_grouping["filter_interaction"],
        ctds_parameters=ctx.args_grouping["parameters"],
        ctd_theme=ctx.args_grouping["theme_selector"],
    )



from typing import List, Literal

import dash_bootstrap_components as dbc
from dash import html
from pydantic import Field

from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._models_utils import _log_call


class Button(VizroBaseModel):
    """Component provided to `Page` to trigger any defined `action` in `Page`.

    Args:
        type (Literal["button"]): Defaults to `"button"`.
        text (str): Text to be displayed on button. Defaults to `"Click me!"`.
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.
    """

    type: Literal["button"] = "button"
    text: str = Field("Click me!", description="Text to be displayed on button.")
    actions: List[Action] = []

    # Re-used validators
    _set_actions = _action_validator_factory("n_clicks")

    @_log_call
    def build(self):
        return html.Div(
            [
                dbc.Button(
                    id=self.id,
                    children=self.text,
                    className="button_primary",
                ),
            ],
            className="button_container",
            id=f"{self.id}_outer",
        )



"""Unit tests for vizro.models.ActionChain."""
import pytest

from vizro.models._action._action import Action
from vizro.models._action._actions_chain import ActionsChain, Trigger, _set_actions
from vizro.models.types import CapturedCallable


@pytest.fixture
def test_trigger():
    return Trigger("component_id", "value")


@pytest.fixture
def test_action(test_action_function):
    return Action(function=test_action_function)


class TestActionsChainInstantiation:
    """Tests model instantiation."""

    def test_create_actions_chain_mandatory_only(self, test_trigger):
        actions_chain = ActionsChain(trigger=test_trigger)

        assert hasattr(actions_chain, "id")
        assert actions_chain.trigger == test_trigger
        assert actions_chain.actions == []

    def test_create_action_chains_mandatory_and_optional(self, test_trigger, test_action):
        actions_chain = ActionsChain(trigger=test_trigger, actions=[test_action], id="actions_chain_id")

        assert actions_chain.id == "actions_chain_id"
        assert actions_chain.trigger == test_trigger
        assert isinstance(actions_chain.actions, list)
        assert actions_chain.actions[0] == test_action


def test_set_actions(test_action):
    result = _set_actions(actions=[test_action], values={"id": "component_id"}, trigger_property="value")
    actions_chain = result[0]
    action = actions_chain.actions[0]

    assert len(result) == 1
    assert isinstance(actions_chain, ActionsChain)
    assert actions_chain.trigger.component_property == "value"
    assert isinstance(action, Action)
    assert isinstance(action.function, CapturedCallable)
    assert action.inputs == []
    assert action.outputs == []



# Redundant aliases here to prevent ruff from removing unused imports.
from typing import Any, Callable, Dict

from vizro.actions._filter_action import _filter
from vizro.actions._on_page_load_action import _on_page_load
from vizro.actions._parameter_action import _parameter
from vizro.actions.export_data_action import export_data
from vizro.actions.filter_interaction_action import filter_interaction

# Please keep alphabetically ordered
__all__ = ["export_data", "filter_interaction"]

# Actions lookup dictionary to facilitate function comparison
action_functions: Dict[Callable[[Any], Dict[str, Any]], str] = {
    export_data.__wrapped__: "export_data",
    filter_interaction.__wrapped__: "filter_interaction",
    _filter.__wrapped__: "filter",
    _parameter.__wrapped__: "parameter",
    _on_page_load.__wrapped__: "on_page_load",
}



from typing import Dict, List, Literal, Optional

from dash import ClientsideFunction, Input, Output, State, clientside_callback, dcc, html
from pydantic import Field, validator

from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._components.form._form_utils import (
    set_default_marks,
    validate_max,
    validate_slider_value,
    validate_step,
)
from vizro.models._models_utils import _log_call


class Slider(VizroBaseModel):
    """Numeric single-selector `Slider`.

    Can be provided to [`Filter`][vizro.models.Filter] or
    [`Parameter`][vizro.models.Parameter]. Based on the underlying
    [`dcc.Slider`](https://dash.plotly.com/dash-core-components/slider).

    Args:
        type (Literal["range_slider"]): Defaults to `"range_slider"`.
        min (Optional[float]): Start value for slider. Defaults to `None`.
        max (Optional[float]): End value for slider. Defaults to `None`.
        step (Optional[float]): Step-size for marks on slider. Defaults to `None`.
        marks (Optional[Dict[float, str]]): Marks to be displayed on slider. Defaults to `None`.
        value (Optional[float]): Default value for slider. Defaults to `None`.
        title (Optional[str]): Title to be displayed. Defaults to `None`.
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.
    """

    type: Literal["slider"] = "slider"
    min: Optional[float] = Field(None, description="Start value for slider.")
    max: Optional[float] = Field(None, description="End value for slider.")
    step: Optional[float] = Field(None, description="Step-size for marks on slider.")
    marks: Optional[Dict[float, str]] = Field(None, description="Marks to be displayed on slider.")
    value: Optional[float] = Field(None, description="Default value for slider.")
    title: Optional[str] = Field(None, description="Title to be displayed.")
    actions: List[Action] = []

    # Re-used validators
    _validate_max = validator("max", allow_reuse=True)(validate_max)
    _validate_value = validator("value", allow_reuse=True)(validate_slider_value)
    _validate_step = validator("step", allow_reuse=True)(validate_step)
    _set_default_marks = validator("marks", allow_reuse=True, always=True)(set_default_marks)
    _set_actions = _action_validator_factory("value")

    @_log_call
    def build(self):
        output = [
            Output(f"{self.id}_text_value", "value"),
            Output(self.id, "value"),
            Output(f"{self.id}_temp_store", "data"),
        ]
        inputs = [
            Input(f"{self.id}_text_value", "value"),
            Input(self.id, "value"),
            State(f"{self.id}_temp_store", "data"),
            State(f"{self.id}_callback_data", "data"),
        ]

        clientside_callback(
            ClientsideFunction(namespace="clientside", function_name="update_slider_values"),
            output=output,
            inputs=inputs,
        )

        return html.Div(
            [
                dcc.Store(
                    f"{self.id}_callback_data",
                    data={
                        "id": self.id,
                        "min": self.min,
                        "max": self.max,
                    },
                ),
                html.P(self.title) if self.title else None,
                html.Div(
                    [
                        dcc.Slider(
                            id=self.id,
                            min=self.min,
                            max=self.max,
                            step=self.step,
                            marks=self.marks,
                            value=self.value or self.min,
                            included=False,
                            persistence=True,
                            className="slider_control" if self.step else "slider_control_no_space",
                        ),
                        dcc.Input(
                            id=f"{self.id}_text_value",
                            type="number",
                            placeholder="end",
                            min=self.min,
                            max=self.max,
                            value=self.value or self.min,
                            persistence=True,
                            className="slider_input_field_right" if self.step else "slider_input_field_no_space_right",
                        ),
                        dcc.Store(id=f"{self.id}_temp_store", storage_type="local"),
                    ],
                    className="slider_inner_container",
                ),
            ],
            className="selector_container",
            id=f"{self.id}_outer",
        )



from typing import Optional

import dash
import dash_bootstrap_components as dbc
from dash import html
from pydantic import validator

from vizro._constants import ACCORDION_TITLE, MODULE_PAGE_404
from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models._navigation.navigation import _validate_pages
from vizro.models.types import NavigationPagesType


class Accordion(VizroBaseModel):
    """Accordion to be used in Navigation Panel of Dashboard.

    Args:
        pages (Optional[NavigationPagesType]): See [NavigationPagesType][vizro.models.types.NavigationPagesType].
            Defaults to `None`.
    """

    pages: Optional[NavigationPagesType] = None

    # validators
    _validate_pages = validator("pages", allow_reuse=True, always=True)(_validate_pages)

    @_log_call
    def build(self):
        if self.pages:
            return self._create_accordion()
        return self._create_default_accordion()

    def _create_accordion_buttons(self, accordion_pages):
        return [
            dbc.Button(
                children=[page["name"]],
                key=page["relative_path"],
                className="accordion-item-button",
                href=page["relative_path"],
            )
            for page in dash.page_registry.values()
            for accordion_page in accordion_pages
            if accordion_page == page["module"] and page["module"] != MODULE_PAGE_404
        ]

    def _create_accordion_item(self, accordion_buttons, title=ACCORDION_TITLE):
        return dbc.AccordionItem(
            children=accordion_buttons,
            title=title.upper(),
            class_name="accordion_item",
        )

    def _create_default_accordion(self):
        accordion_buttons = [
            dbc.Button(
                children=[page["name"]],
                key=page["relative_path"],
                className="accordion-item-button",
                href=page["relative_path"],
            )
            for page in dash.page_registry.values()
            if page["module"] != MODULE_PAGE_404
        ]

        accordion_items = [self._create_accordion_item(accordion_buttons=accordion_buttons)]

        # Don't create accordion navigation if there is only one page and one accordion item
        if len(accordion_buttons) == len(accordion_items) == 1:
            return None

        return html.Div(
            children=[
                dbc.Accordion(
                    id=self.id,
                    children=accordion_items,
                    class_name="accordion",
                    persistence=True,
                    persistence_type="session",
                ),
                html.Div(className="keyline"),
            ],
            className="nav_panel",
            id=f"{self.id}_outer",
        )

    def _create_accordion(self):
        accordion_items = []

        if isinstance(self.pages, dict):
            for title, accordion_pages in self.pages.items():
                accordion_buttons = self._create_accordion_buttons(accordion_pages=accordion_pages)
                accordion_items.append(self._create_accordion_item(accordion_buttons=accordion_buttons, title=title))

        if isinstance(self.pages, list):
            accordion_buttons = self._create_accordion_buttons(accordion_pages=self.pages)
            accordion_items.append(self._create_accordion_item(accordion_buttons=accordion_buttons))

        if len(accordion_buttons) == len(accordion_items) == 1:
            return None

        return html.Div(
            children=[
                dbc.Accordion(
                    id=self.id,
                    children=accordion_items,
                    class_name="accordion",
                    persistence=True,
                    persistence_type="session",
                ),
                html.Div(className="keyline"),
            ],
            className="nav_panel",
            id=f"{self.id}_outer",
        )
